// testing tool for maniaplanet_telemetry

#include <windows.h>

#include <stdio.h>
#include <math.h>
#include <stdint.h>

#include <SDL2/SDL.h>
#include <SDL2/SDL_opengl.h>
#undef main

#include <imgui.h>
#include "imgui_impl_sdl.h"

#include "maniaplanet_telemetry.h"

static int rad2deg(float r) {  return int(r*180/3.14152f); }

static NManiaPlanet::Vec3 RotateInv(NManiaPlanet::Quat r, NManiaPlanet::Vec3 v)
{
    float t2 =  -r.w*r.x;
    float t3 =  -r.w*r.y;
    float t4 =  -r.w*r.z;
    float t5 =  -r.x*r.x;
    float t6 =   r.x*r.y;
    float t7 =   r.x*r.z;
    float t8 =  -r.y*r.y;
    float t9 =   r.y*r.z;
    float t10 = -r.z*r.z;

    NManiaPlanet::Vec3 Res;
    Res.x = 2*( (t8 + t10)*v.x + (t6 -  t4)*v.y + (t3 + t7)*v.z ) + v.x;
    Res.y = 2*( (t4 +  t6)*v.x + (t5 + t10)*v.y + (t9 - t2)*v.z ) + v.y;
    Res.z = 2*( (t7 -  t3)*v.x + (t2 +  t9)*v.y + (t5 + t8)*v.z ) + v.z;
    return Res;
}

int main(int, char**)
{
	// Setup SDL
	if (SDL_Init(SDL_INIT_EVERYTHING) != 0)
		return -1;

	// Setup window
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
	SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 2);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 2);
	SDL_DisplayMode current;
	SDL_GetCurrentDisplayMode(0, &current);
    SDL_Window *sdl_window = SDL_CreateWindow("testbed", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, 1280, 720, SDL_WINDOW_OPENGL|SDL_WINDOW_RESIZABLE);
	SDL_GLContext glcontext = SDL_GL_CreateContext(sdl_window);

	// Setup ImGui binding
	ImGui_ImplSdl_Init(sdl_window);

    HANDLE hMapFile = NULL;
    void* pBufView = NULL;

    using namespace NManiaPlanet;
    const volatile STelemetry* Shared = NULL;

	// Main loop
	bool done = false;
	while (!done)
	{
		SDL_Event event;
		while (SDL_PollEvent(&event))
		{
			ImGui_ImplSdl_ProcessEvent(&event);
			if (event.type == SDL_QUIT)
				done = true;
		}
        ImGui_ImplSdl_NewFrame(sdl_window);


        ImGui::SetNextWindowPos(ImVec2(10, 10));
        ImGui::SetNextWindowSize(ImVec2(500, 700));
        ImGui::Begin("win1", NULL, ImGuiWindowFlags_NoTitleBar|ImGuiWindowFlags_NoResize|ImGuiWindowFlags_NoMove);

        // top bar.
        {
            if (ImGui::Button("Quit"))
                break;
        }

        ImGui::Spacing();	// -----------------

        if (Shared == NULL)
        {
            ImGui::Text("Waiting for the game...");

            if (!hMapFile)
                hMapFile = OpenFileMapping( FILE_MAP_ALL_ACCESS,         // read/write access
                                            FALSE,                       // do not inherit the name
                                            "ManiaPlanet_Telemetry");    // name of mapping object
            if (hMapFile) {
                if (!pBufView)
                    pBufView = (void*) MapViewOfFile(   hMapFile,   // handle to map object
                                                    FILE_MAP_ALL_ACCESS, // read/write permission
                                                    0,
                                                    0,
                                                    4096);
            }
            Shared = (const STelemetry*)pBufView;
        }
        else
        {
            STelemetry S;
            for (;;) {
                Nat32 Before = Shared->UpdateNumber;
                memcpy(&S, (const STelemetry*)Shared, sizeof(S));
                Nat32 After = Shared->UpdateNumber;
                if (Before == After)
                    break;
                else
                    continue; // reading while the game is changing the values.. retry.
            }

            Vec3 LocalVel = RotateInv(S.Object.Rotation, S.Object.Velocity);

            if (ImGui::CollapsingHeader("Header", NULL, true, true))
            {
                ImGui::Columns(2, "Header");
                ImGui::Text("Magic");			ImGui::NextColumn();	ImGui::Text(S.Header.Magic);            ImGui::NextColumn();
                ImGui::Text("Version");			ImGui::NextColumn();	ImGui::Text("%d (%d)", S.Header.Version, ECurVersion);    ImGui::NextColumn();
                ImGui::Text("Size");			ImGui::NextColumn();	ImGui::Text("%d (%d)", S.Header.Size, sizeof(S));  ImGui::NextColumn();
                ImGui::Columns(1);
            }
            if (ImGui::CollapsingHeader("Game", NULL, true, true))
            {
                static const char* StateNames[] = { "Starting", "Menus", "Running", "Paused" };
                ImGui::Columns(2, "Game");
                ImGui::Text("State");			ImGui::NextColumn();	ImGui::Combo("##State", (int*)&S.Game.State, StateNames, (int)(sizeof(StateNames)/sizeof(StateNames[0])));  ImGui::NextColumn();
                ImGui::Text("GameplayVariant");	ImGui::NextColumn();	ImGui::Text(S.Game.GameplayVariant);    ImGui::NextColumn();
                ImGui::Text("MapName");			ImGui::NextColumn();	ImGui::Text(S.Game.MapName);            ImGui::NextColumn();
                ImGui::Text("MapId");			ImGui::NextColumn();	ImGui::Text(S.Game.MapId);              ImGui::NextColumn();
                ImGui::Columns(1);
            }
            if (ImGui::CollapsingHeader("Race", NULL, true, true))
            {
                ImGui::Columns(2, "Race");
                static const char* StateNames[] = { "BeforeStart", "Running", "Finished" };
                ImGui::Text("State");               ImGui::NextColumn();	ImGui::Combo("##State", (int*)&S.Race.State, StateNames, (int)(sizeof(StateNames)/sizeof(StateNames[0])));  ImGui::NextColumn();
                ImGui::Text("RaceTime");            ImGui::NextColumn();	ImGui::Text("%d", S.Race.Time);             ImGui::NextColumn();
                ImGui::Text("NbRespawns");          ImGui::NextColumn();	ImGui::Text("%d", S.Race.NbRespawns);       ImGui::NextColumn();
                ImGui::Text("NbCheckpoints");       ImGui::NextColumn();	ImGui::Text("%d", S.Race.NbCheckpoints);    ImGui::NextColumn();
                ImGui::Columns(1);
            }
            if (ImGui::CollapsingHeader("Object", NULL, true, true))
            {
                ImGui::Columns(2, "Object");
                ImGui::Text("Timestamp");           ImGui::NextColumn();	ImGui::Text("%d", S.Object.Timestamp);    ImGui::NextColumn();
                ImGui::Text("DiscontinuityCount");	ImGui::NextColumn();	ImGui::Text("%d", S.Object.DiscontinuityCount);            ImGui::NextColumn();
                ImGui::Text("Translation");         ImGui::NextColumn();	ImGui::Text("%.1f %.1f %.1f", S.Object.Translation.x, S.Object.Translation.y, S.Object.Translation.z);      ImGui::NextColumn();
                ImGui::Text("Rotation(Quat)");      ImGui::NextColumn();	ImGui::Text("%.1f %.1f %.1f %.1f", S.Object.Rotation.x, S.Object.Rotation.y, S.Object.Rotation.z, S.Object.Rotation.w);      ImGui::NextColumn();
                ImGui::Text("WorldVelocity");       ImGui::NextColumn();	ImGui::Text("%.1f %.1f %.1f", S.Object.Velocity.x, S.Object.Velocity.y, S.Object.Velocity.z);               ImGui::NextColumn();
                ImGui::Text("LocalVelocity");       ImGui::NextColumn();	ImGui::Text("%.1f %.1f %.1f", LocalVel.x, LocalVel.y, LocalVel.z);    ImGui::NextColumn();
                ImGui::Text("LatestGroundContact"); ImGui::NextColumn();	ImGui::Text("%d", S.Object.LatestStableGroundContactTime);            ImGui::NextColumn();
                ImGui::Columns(1);
            }
            if (ImGui::CollapsingHeader("Device", NULL, true, true))
            {
                ImGui::Columns(2, "Object");
                ImGui::Text("Euler");               ImGui::NextColumn();	ImGui::Text("%d %d %d", rad2deg(S.Device.Euler.x), rad2deg(S.Device.Euler.y), rad2deg(S.Device.Euler.z));   ImGui::NextColumn();
                ImGui::Text("CenteredAltitude");    ImGui::NextColumn();	ImGui::Text("%.1f", S.Device.CenteredAltitude);    ImGui::NextColumn();
                ImGui::Text("CenteredYaw");         ImGui::NextColumn();	ImGui::Text("%d", rad2deg(S.Device.CenteredYaw));            ImGui::NextColumn();
                ImGui::Columns(1);
            }
            ImGui::End();

            ImGui::SetNextWindowPos(ImVec2(520, 10));
            ImGui::SetNextWindowSize(ImVec2(500, 700));
            ImGui::Begin("win2", NULL, ImGuiWindowFlags_NoTitleBar|ImGuiWindowFlags_NoResize|ImGuiWindowFlags_NoMove);

            if (ImGui::CollapsingHeader("Vehicle", NULL, true, true))
            {
                ImGui::Columns(2, "Vehicle");
                ImGui::Text("Timestamp");           ImGui::NextColumn();	ImGui::Text("%d", S.Vehicle.Timestamp);     ImGui::NextColumn();
                ImGui::Text("SpeedMeter");          ImGui::NextColumn();	ImGui::Text("%d km/h", S.Vehicle.SpeedMeter);    ImGui::NextColumn();
                ImGui::Separator();
                ImGui::Text("InputSteer");          ImGui::NextColumn();	ImGui::Text("%f", S.Vehicle.InputSteer);    ImGui::NextColumn();
                ImGui::Text("InputGasPedal");       ImGui::NextColumn();	ImGui::Text("%f", S.Vehicle.InputGasPedal); ImGui::NextColumn();
                ImGui::Text("InputIsBraking");      ImGui::NextColumn();	ImGui::Checkbox("##Brake", (bool*)&S.Vehicle.InputIsBraking); ImGui::NextColumn();
                ImGui::Text("InputIsHorn");         ImGui::NextColumn();	ImGui::Checkbox("##Horn", (bool*)&S.Vehicle.InputIsHorn); ImGui::NextColumn();
                ImGui::Separator();
                ImGui::Text("EngineRpm");           ImGui::NextColumn();	ImGui::Text("%f", S.Vehicle.EngineRpm);     ImGui::NextColumn();
                ImGui::Text("EngineCurGear");       ImGui::NextColumn();	ImGui::Text("%d", S.Vehicle.EngineCurGear);    ImGui::NextColumn();
                ImGui::Text("EngineTurboRatio");    ImGui::NextColumn();	ImGui::Text("%f", S.Vehicle.EngineTurboRatio);    ImGui::NextColumn();
                ImGui::Text("EngineFreeWheeling");  ImGui::NextColumn();	ImGui::Checkbox("##EngOff", (bool*)&S.Vehicle.EngineFreeWheeling); ImGui::NextColumn();
                ImGui::Separator();
                ImGui::Text("WheelsContact");       ImGui::NextColumn();	ImGui::Text("%d %d %d %d", S.Vehicle.WheelsIsGroundContact[0], S.Vehicle.WheelsIsGroundContact[1], S.Vehicle.WheelsIsGroundContact[2], S.Vehicle.WheelsIsGroundContact[3]);     ImGui::NextColumn();
                ImGui::Text("WheelsSliping");       ImGui::NextColumn();	ImGui::Text("%d %d %d %d", S.Vehicle.WheelsIsSliping[0], S.Vehicle.WheelsIsSliping[1], S.Vehicle.WheelsIsSliping[2], S.Vehicle.WheelsIsSliping[3]);    ImGui::NextColumn();
                ImGui::Text("WheelsDamperLen");     ImGui::NextColumn();	ImGui::Text("%.4f %.4f %.4f %.4f", S.Vehicle.WheelsDamperLen[0], S.Vehicle.WheelsDamperLen[1], S.Vehicle.WheelsDamperLen[2], S.Vehicle.WheelsDamperLen[3]);    ImGui::NextColumn();
                ImGui::Separator();
                ImGui::Text("InWater");             ImGui::NextColumn();	ImGui::Checkbox("##InWater", (bool*)&S.Vehicle.IsInWater); ImGui::NextColumn();
                ImGui::Text("Sparkling");           ImGui::NextColumn();	ImGui::Checkbox("##IsSparkling", (bool*)&S.Vehicle.IsSparkling); ImGui::NextColumn();
                ImGui::Text("LightTrails");         ImGui::NextColumn();	ImGui::Checkbox("##IsLightTrails", (bool*)&S.Vehicle.IsLightTrails); ImGui::NextColumn();
                ImGui::Text("LightsOn");            ImGui::NextColumn();	ImGui::Checkbox("##LightsOn", (bool*)&S.Vehicle.IsLightsOn); ImGui::NextColumn();
                ImGui::Text("Flying");              ImGui::NextColumn();	ImGui::Checkbox("##Flying", (bool*)&S.Vehicle.IsFlying); ImGui::NextColumn();
                ImGui::Text("Rumble");              ImGui::NextColumn();	ImGui::Text("%f", S.Vehicle.RumbleIntensity);    ImGui::NextColumn();
                ImGui::Columns(1);
            }
        }

        ImGui::End();

		// Rendering
		ImVec4 clear_color = ImColor(114, 144, 154);
		glViewport(0, 0, (int)ImGui::GetIO().DisplaySize.x, (int)ImGui::GetIO().DisplaySize.y);
		glClearColor(clear_color.x, clear_color.y, clear_color.z, clear_color.w);
		glClear(GL_COLOR_BUFFER_BIT);
		ImGui::Render();
		SDL_GL_SwapWindow(sdl_window);
	}


    // Cleanup

    if (pBufView)
        UnmapViewOfFile(pBufView);
    if (hMapFile)
        CloseHandle(hMapFile);

	ImGui_ImplSdl_Shutdown();
	SDL_GL_DeleteContext(glcontext);
	SDL_DestroyWindow(sdl_window);
	SDL_Quit();

	return 0;
}
